import state
import gameobj
import pyglet
import graph
import random
import font
import event
import os
import data

DEBUG = True

image_filepath = data.filepath('img')

menu_bg = pyglet.image.load(os.path.join(image_filepath, 'menu_bg.png'))
game_bg = pyglet.image.load(os.path.join(image_filepath, 'game_bg.png'))

button_new_game_img = pyglet.image.load(os.path.join(image_filepath, 'button_new_game.png'))
button_exit_img = pyglet.image.load(os.path.join(image_filepath, 'exit_game.png'))

button_nest_unavail_img = pyglet.image.load(os.path.join(image_filepath, 'button_nest_unavail.png'))
button_nest_avail_red_img = pyglet.image.load(os.path.join(image_filepath, 'button_nest_avail_red.png'))
button_nest_avail_blue_img = pyglet.image.load(os.path.join(image_filepath, 'button_nest_avail_blue.png'))

class MenuState(state.State):
	
	NAME = 'MENU'
	
	def __init__(self):
		super(MenuState, self).__init__(self.__class__.NAME)
		
	def init(self):
		print '%s init' % self.NAME
	
	def init_display(self, width, height):
		print '%s init display' % self.NAME
	
	def cleanup(self):
		print '%s cleanup' % self.NAME
	
	def enter(self):
		print '%s enter' % self.NAME
		
		self.new_game_sprite = pyglet.sprite.Sprite(button_new_game_img)
		self.new_game_sprite.x = (800. // 2) - (self.new_game_sprite.width // 2)
		self.new_game_sprite.y = 600 - self.new_game_sprite.height - 200
		
		self.exit_game_sprite = pyglet.sprite.Sprite(button_exit_img)
		self.exit_game_sprite.x = (800. // 2) - (self.exit_game_sprite.width // 2)
		self.exit_game_sprite.y = 600 - self.exit_game_sprite.height - 300
	
	def exit(self):
		print '%s exit' % self.NAME
	
	def pause(self):
		print '%s pause' % self.NAME
	
	def unpause(self):
		print '%s unpause' % self.NAME
		
		pyglet.gl.glClearColor(.2, .2, .2, 0)
	
	def on_key_press(self, symbol, modifiers):
		#if symbol == pyglet.window.key.ENTER:
		#	self.manager.push_state(GameState.NAME)
		pass
	
	def on_mouse_press(self, x, y, button, modifiers):
		if self.new_game_sprite.x <= x and self.new_game_sprite.x + self.new_game_sprite.width >= x:
			if self.new_game_sprite.y <= y and self.new_game_sprite.y + self.new_game_sprite.height >= y:
				self.manager.push_state(GameState.NAME)
			
		if self.exit_game_sprite.x <= x and self.exit_game_sprite.x + self.exit_game_sprite.width >= x:
			if self.exit_game_sprite.y <= y and self.exit_game_sprite.y + self.exit_game_sprite.height >= y:
				pyglet.app.exit()
	
	def update(self, dt):
		pass
	
	def draw(self):
		menu_bg.blit(0, 0)
		
		self.new_game_sprite.draw()
		self.exit_game_sprite.draw()

class GameState(state.State):
	
	NAME = 'GAME'
	
	def __init__(self):
		super(GameState, self).__init__(self.__class__.NAME)
		
	def init(self):
		print '%s init' % self.NAME
	
	def init_display(self, width, height):
		self.display_size = (width, height)
	
	def cleanup(self):
		print '%s cleanup' % self.NAME
	
	def enter(self):
		print '%s enter' % self.NAME

		self.human_player = gameobj.PLAYER1

		self.manager.model = self._build_world(self.human_player)
		self.world_event_handler = GameStateWorldEventHandler(self)
		
		if DEBUG:
			for l in self.manager.model.locations:
				l.label = pyglet.text.Label('%s' % l.name,
							font_name='Arial',
							font_size=12,
							x=l.pos.x - 25, y=l.pos.y,
							anchor_x='right', anchor_y='center')
		
		self.manager.model.init()
		self.world_event_handler.init()
		
		self.manager.model.register_handler(self.world_event_handler)
		
		self.p1_score_label = pyglet.text.Label('',
						  font_name='Armor Piercing 2.0 BB',
						  font_size=18,
						  x=230, y=self.display_size[1] - 10,
						  anchor_x='left', anchor_y='top')
						  
		self.p2_score_label = pyglet.text.Label('',
						  font_name='Armor Piercing 2.0 BB',
						  font_size=18,
						  x=self.display_size[0] - 250, y=self.display_size[1] - 10,
						  anchor_x='left', anchor_y='top')
						  
		self.paused_label = pyglet.text.Label('Paused',
							font_name='Armor Piercing 2.0 BB',
							font_size=30,
							x=self.display_size[0] // 2, y=self.display_size[1] - 10,
							anchor_x='center', anchor_y='top',
							color=(255, 255, 255, 255))
							
		self.hint_label = pyglet.text.Label('Hint',
							font_name='Armor Piercing 2.0 BB',
							font_size=16,
							x=5, y=5,
							anchor_x='left', anchor_y='bottom',
							color=(255, 255, 255, 255))
		self.hint = ''
	
		self.labels = []
		
		self.paused = False
		
		self.hover_node = None
		
		self.nest_unavail_sprite = pyglet.sprite.Sprite(button_nest_unavail_img)		
		if self.human_player == gameobj.PLAYER1:
			self.nest_avail_sprite = pyglet.sprite.Sprite(button_nest_avail_red_img)
			self.nest_avail_sprite.x = 5
			self.nest_avail_sprite.y = 600 - self.nest_avail_sprite.height - 5
			
			self.nest_unavail_sprite.x = 5
			self.nest_unavail_sprite.y = 600 - self.nest_avail_sprite.height - 5
				
		else:
			self.nest_avail_sprite = pyglet.sprite.Sprite(button_nest_avail_blue_img)
			self.nest_avail_sprite.x = 800 - self.nest_avail_sprite.width - 5
			self.nest_avail_sprite.y = 600 - self.nest_avail_sprite.height - 5
			
			self.nest_unavail_sprite.x = 800 - self.nest_avail_sprite.width - 5
			self.nest_unavail_sprite.y = 600 - self.nest_avail_sprite.height - 5
		
	def exit(self):
		print '%s exit' % self.NAME
		
		self.manager.model.cleanup()
		self.world_event_handler.cleanup()
		
		self.manager.model.deregister_handler(self.world_event_handler)
		
		#del self.manager.model
		#del self.world_event_handler
		
		del self.labels
	
	def pause(self):
		print '%s pause' % self.NAME
	
	def unpause(self):
		print '%s unpause' % self.NAME
		
		pyglet.gl.glClearColor(0.1, 0.1, 0.1, 0)
	
	def on_key_press(self, symbol, modifiers):
		if symbol == pyglet.window.key.ESCAPE:
			self.manager.pop_state()

		world = self.manager.model
		if not world.game_over:					
			if symbol == pyglet.window.key.P:
				self.paused = not self.paused
			
			if not self.paused:
				if symbol == pyglet.window.key.N:
					
						hero = world.player_hero(self.human_player)
						if hero:
							if hero.can_nest():
								hero.create_nest()
	
	def on_mouse_press(self, x, y, button, modifiers):
		world = self.manager.model
		if not world.game_over and not self.paused:
			hero = world.player_hero(self.human_player)
			if hero:			
				for action in hero.action_stack:
					print action
					
				for node in world.locations:
					if gameobj.intersects_point(node, x, y):
						if button == pyglet.window.mouse.LEFT:
							hero.move_to(node)
							break
							
				if self.nest_avail_sprite.x <= x and self.nest_avail_sprite.x + self.nest_avail_sprite.width >= x:
					if self.nest_avail_sprite.y <= y and self.nest_avail_sprite.y + self.nest_avail_sprite.height >= y:
						if hero.can_nest():
							hero.create_nest()
	
	def on_mouse_motion(self, x, y, dx, dy):
		self.hint = ''
		self.hover_node = None
		world = self.manager.model
		if not world.game_over:
			for node in world.locations:
				if gameobj.intersects_point(node, x, y):	
					self.hover_node = node
					self.hint = 'Click to move hero here'
					break

			if self.nest_avail_sprite.x <= x and self.nest_avail_sprite.x + self.nest_avail_sprite.width >= x:
				if self.nest_avail_sprite.y <= y and self.nest_avail_sprite.y + self.nest_avail_sprite.height >= y:
					self.hint = 'Orders your hero to build a nest. Nest cost = %s points' % gameobj.NestEntity.get_point_cost()
	
	def cleanup(self):
		pass
		
	def update(self, dt):
		if self.paused:
			return
			
		self.manager.model.update(dt)
		
		rem_list = []
		for label in self.labels:
			label.update(dt)
			
			if label.tl >= label.ttl:
				rem_list.append(label)
				
			label.label.x += label.pos_mod[0]
			label.label.y += label.pos_mod[1]
			color = list(label.label.color)
			
			if color[3] >= label.alpha_decay:
				color[3] -= label.alpha_decay
			else:
				color[3] = 0
			label.label.color = tuple(color)

		for label in rem_list:
			self.labels.remove(label)
			label.cleanup()

	def draw(self):
		game_bg.blit(0, 0)
		
		world = self.manager.model
		
		visited_nodes = set()
		for node in world.locations:
			pyglet.gl.glLineWidth(2.5) 
			for other_node in world.locations:
				if node == other_node:
					continue
				if other_node in visited_nodes:
					continue	
				
				if world.graph.connected(node, other_node):
					pyglet.graphics.draw(2, pyglet.gl.GL_LINES,
						('v2f', (node.pos.x, node.pos.y, other_node.pos.x, other_node.pos.y))
					)
					
				if DEBUG:
					node.label.draw()
			
			visited_nodes.add(node)
		
		if self.hover_node:
			pyglet.graphics.draw(4, pyglet.gl.GL_LINE_LOOP,
				('v2f', (self.hover_node.pos.x - 20, self.hover_node.pos.y - 20, 
						self.hover_node.pos.x - 20, self.hover_node.pos.y + 20,
						self.hover_node.pos.x + 20, self.hover_node.pos.y + 20,
						self.hover_node.pos.x + 20, self.hover_node.pos.y - 20))
			)
		
		for node in world.locations:
			node.sprite.draw()
		
		heroes = []
		minions = []
		nests = []
		bases = []
		for entity in world.entities:
			if isinstance(entity, gameobj.HeroEntity):
				heroes.append(entity)
			if isinstance(entity, gameobj.MinionEntity):
				minions.append(entity)
			if isinstance(entity, gameobj.NestEntity):
				nests.append(entity)
			if isinstance(entity, gameobj.BaseEntity):
				bases.append(entity)
		
		if not world.game_over:
			for e in bases:
				e.sprite.draw()
		for e in nests:
			e.sprite.draw()
		for e in minions:
			e.sprite.draw()
		for e in heroes:
			e.sprite.draw()
		
		#world.batch.draw()	
		
		if not world.game_over:
			pyglet.gl.glLineWidth(1)
			for entity in world.entities:
				if isinstance(entity, gameobj.HeroEntity) or isinstance(entity, gameobj.BaseEntity) or isinstance(entity, gameobj.NestEntity):
					width = 30.
					health_perc = (float(entity.health) / entity.health_max)
					rv = int(255 * (1 - health_perc))
					gv = int(255 * (health_perc))
					fill_width = width * health_perc
					height = 3.
					bottom = entity.sprite.height // 2 + 5
					pyglet.graphics.draw(4, pyglet.gl.GL_LINE_LOOP,
						('v2f', (entity.pos.x - width // 2, entity.pos.y + bottom,
								 entity.pos.x - width // 2, entity.pos.y + bottom + height,
								 entity.pos.x + width // 2, entity.pos.y + bottom + height,
								 entity.pos.x + width // 2, entity.pos.y + bottom),
						),
						('c3B', (200, 200, 200, 
								 200, 200, 200,
								 200, 200, 200,
								 200, 200, 200,)
						)
					)
					pyglet.graphics.draw(4, pyglet.gl.GL_QUADS,
						('v2f', (entity.pos.x - width // 2, entity.pos.y + bottom,
								 entity.pos.x - width // 2, entity.pos.y + bottom + height,
								 entity.pos.x - width // 2 + fill_width, entity.pos.y + bottom + height,
								 entity.pos.x - width // 2 + fill_width, entity.pos.y + bottom),
						),
						('c3B', (rv, gv, 0, 
								 rv, gv, 0,
								 rv, gv, 0,
								 rv, gv, 0,)
						)
					)
					
				if isinstance(entity, gameobj.HeroEntity):
					for action in entity.action_stack:
						if isinstance(action, gameobj.SpawnNestEntityAction):
							total_time = float(gameobj.SpawnNestEntityAction.get_spawn_wait_ticks())
							countdown = float(action.spawn_countdown)
							progress = (total_time - countdown) / (total_time / 100) / 100
							
							bottom += height + 1
							fill_width = width * progress
							
							pyglet.graphics.draw(4, pyglet.gl.GL_QUADS,
								('v2f', (entity.pos.x - width // 2, entity.pos.y + bottom,
										 entity.pos.x - width // 2, entity.pos.y + bottom + height,
										 entity.pos.x - width // 2 + fill_width, entity.pos.y + bottom + height,
										 entity.pos.x - width // 2 + fill_width, entity.pos.y + bottom),
								),
								('c3B', (255, 255, 0, 
										 255, 255, 0,
										 255, 255, 0,
										 255, 255, 0,)
								)
							)
				
			attribs = world.player_attributes(gameobj.PLAYER1)		
			self.p1_score_label.text = '%s ' % attribs.points
			self.p1_score_label.draw()
			
			attribs = world.player_attributes(gameobj.PLAYER2)		
			self.p2_score_label.text = '%s ' % attribs.points
			self.p2_score_label.draw()
			
			self.hint_label.text = self.hint
			self.hint_label.draw()
			
			for label in self.labels:
				label.label.draw()
			
			hero = world.player_hero(self.human_player)
			if hero and hero.can_nest():
				self.nest_avail_sprite.draw()
			else:
				self.nest_unavail_sprite.draw()			
			
		if world.game_over:
			if self.human_player == world.winning_player:
				self.paused_label.text = 'Victory!'
			else:
				self.paused_label.text = 'Defeat'
			
			self.paused_label.draw()
			
		elif self.paused:
			self.paused_label.draw()

	def _add_label(self, label):
		self.labels.append(label)
		label.init()

	def _build_world(self, human_player):
		width = 800
		height = 600
					
		tile_size = 100
		tile_padding = 20
					
		cols = width // tile_size
		col_max_tiles = height // tile_size

		x_offset = 0
		y_offset = 0
					
		if cols % 2 == 0:
			cols -= 1
			x_offset = tile_size // 2
					
		mid_col = cols // 2

		col_tile_count = []

		last = 0
		for i in xrange(1, mid_col+2):
			col_tile_count.append(i)
			last = i

		for i in xrange(last - 1, 0, -1):
			col_tile_count.append(i)

		nodes_count = 0
		for i, value in enumerate(col_tile_count):
			if value > col_max_tiles:
				col_tile_count[i] = col_max_tiles
				nodes_count += col_max_tiles
			else:
				nodes_count += col_tile_count[i]

		print 'cols: %s' % cols
		print col_tile_count
		print 'total nodes: %s' % nodes_count

		nodes = []
		entities = []
		col_nodes = [[] for _ in xrange(cols)]
		for i, value in enumerate(col_tile_count):
			stack_height = value * tile_size
			offset = (height - stack_height) // 2
			
			for n in xrange(value):
				ny = offset + (n  * tile_size) + tile_size // 2
				nx = (i * tile_size) + tile_size // 2
				
				node = None
				if len(nodes) == 0:
					node = gameobj.BaseLocation(len(nodes), nx + x_offset, ny, gameobj.PLAYER1)
					
					hero = gameobj.HeroEntity('', node, gameobj.PLAYER1)
					if human_player != gameobj.PLAYER1:
						hero.action_strategy = gameobj.HeroActionStrategy(hero)
					entities.append(hero)
					
					base = gameobj.BaseEntity(len(nodes), node, gameobj.PLAYER1)
					entities.append(base)						
				elif len(nodes) == nodes_count - 1:
					node = gameobj.BaseLocation(len(nodes), nx + x_offset, ny, gameobj.PLAYER2)
					
					hero = gameobj.HeroEntity('', node, gameobj.PLAYER2)
					if human_player != gameobj.PLAYER2:
						hero.action_strategy = gameobj.HeroActionStrategy(hero)
					entities.append(hero)
					
					base = gameobj.BaseEntity(len(nodes), node, gameobj.PLAYER2)
					entities.append(base)						
				else:
					node = gameobj.SpawnLocation(len(nodes), nx + x_offset, ny)
					
				nodes.append(node)
				col_nodes[i].append(node)
				
		g = graph.Graph(nodes)

		# connections
		for curr_col_idx in xrange(len(col_nodes)):
			if curr_col_idx != 0:
				prev_col_idx = curr_col_idx - 1
				prev_col = col_nodes[prev_col_idx]
				curr_col = col_nodes[curr_col_idx]
				
				if len(prev_col) == len(curr_col) - 1:
					for n_idx, n in enumerate(prev_col):
						if curr_col_idx % 2 == 0:
							g.connect(n, curr_col[n_idx])
							g.connect(n, curr_col[n_idx + 1], 1)
						else:
							g.connect(n, curr_col[n_idx], 1)
							g.connect(n, curr_col[n_idx + 1])
						
				if len(prev_col) == len(curr_col):
					for n_idx, n in enumerate(prev_col):
						g.connect(n, curr_col[n_idx])
						
				if len(prev_col) == len(curr_col) + 1:
					for n_idx, n in enumerate(curr_col):
						if curr_col_idx % 2 == 0:						
							g.connect(n, prev_col[n_idx], 1)
							g.connect(n, prev_col[n_idx + 1])
						else:
							g.connect(n, prev_col[n_idx])
							g.connect(n, prev_col[n_idx + 1], 1)
						
				for n_idx, n in enumerate(curr_col):
					if n_idx < len(curr_col) - 1:
						g.connect(n, curr_col[n_idx + 1])

		world = gameobj.World(g, gameobj.PLAYER1, human_player, entities)
		
		return world

class GameLabel(object):
	def __init__(self, label, ttl=100, pos_mod=(0, 0), alpha_decay=1):
		self.label = label
		self.ttl = ttl
		self.tl = 0	
		self.pos_mod = pos_mod
		self.alpha_decay = alpha_decay
	
	def init(self):
		pass
		
	def cleanup(self):
		pass
		
	def update(self, dt):
		self.tl += 1

class GameStateWorldEventHandler(event.EventHandler):
	def __init__(self, state):
		super(GameStateWorldEventHandler, self).__init__()
		self.state = state
	
	def init(self):
		pass
		
	def cleanup(self):
		pass
		
	def handle(self, event):
		if event.name == 'POINT_AWARD':
			label = pyglet.text.Label(text='+%s' % event.data, 
				font_name='Arial', 
				font_size=12., 
				bold=True, 
				italic=False, 
				color=gameobj.player_color(event.target), 
				x=event.source.pos.x, 
				y=event.source.pos.y, 
				width=None, 
				height=None, 
				anchor_x='left', 
				anchor_y='baseline', 
				halign='left', 
				multiline=False, 
				dpi=None, 
				batch=None, 
				group=None)
			
			self.state._add_label(GameLabel(label, ttl=130, pos_mod=(0, 0.3), alpha_decay=2))



