import mathutil
from priodict import priorityDictionary
import math

class Node(object):
	def __init__(self, name, x, y):
		self.name = name
		self.pos = mathutil.Vector2D(x, y)
	
	def __str__(self):
		return '%s' % self.name

class Graph(object):
	def __init__(self, nodes):
		self._dict = {}
		
		for node in nodes:
			self._dict[node] = {}
			
		self._paths = {}
		self._path_cache = None
			
	def connect(self, n1, n2, add_weight=0):
		distance = n1.pos.subtract(n2.pos).mag + add_weight
		
		self._dict[n1][n2] = distance
		self._dict[n2][n1] = distance
	
	def connected(self, n1, n2):
		try:
			if n2 in self._dict[n1].keys():
				if n1 in self._dict[n2].keys():
					return True
		except KeyError:
			pass
		return False
	
	def get_node_connections(self, node):
		return self._dict[node].keys()
	
	def get_shortest_path(self,start,end):
		if self._path_cache:
			try:
				path = self._path_cache[start][end]
				return path
			except KeyError:
				pass	
		
		final_distances,predecessors = self._djk(start,end)
		path = []
		while 1:
			path.append(end)
			if end == start: break
			end = predecessors[end]
		path.reverse()
		return path
	
	def path_distance(self, path):
		total = 0.
		for i, n in enumerate(path):
			if i < len(path) - 1:
				nn = path[i + 1]
				total += self._dict[n][nn]
		return total
					
	def cache_paths(self):
		self._path_cache = {}
		
		for start in self._dict.keys():
			final_distances, predecessors = self._djk(start)
			for end in self._dict.keys():
				if end == start:
					continue
				path = []
				step = end
				while 1:
					path.append(step)
					if step == start: break
					step = predecessors[step]
				path.reverse()
				self._path_cache[start] = {end: path}
	
	def _djk(self,start,end=None):
		final_distances = {}	# dictionary of final distances
		predecessors = {}	# dictionary of predecessors
		estimated_distances = priorityDictionary()   # est.dist. of non-final vert.
		estimated_distances[start] = 0

		for vertex in estimated_distances:
			final_distances[vertex] = estimated_distances[vertex]
			if vertex == end: break

			for edge in self._dict[vertex]:
				path_distance = final_distances[vertex] + self._dict[vertex][edge]
				if edge in final_distances:
					if path_distance < final_distances[edge]:
						raise ValueError, \
	  "Dijkstra: found better path to already-final vertex"
				elif edge not in estimated_distances or path_distance < estimated_distances[edge]:
					estimated_distances[edge] = path_distance
					predecessors[edge] = vertex

		return (final_distances,predecessors)
